import { compact } from '../../util/array.js';
import { getPackageNameFromFilePath, getPackageNameFromModuleSpecifier } from '../../util/modules.js';
import { basename, isInternal, dirname, toAbsolute, isAbsolute } from '../../util/path.js';
import { load } from '../../util/plugin.js';
import { _resolve } from '../../util/require.js';
import { getDependenciesFromConfig } from '../babel/index.js';
import { fallback } from './fallback.js';
import { PACKAGE_JSON_PATH } from './index.js';
const getDependencies = (config) => {
    const extendsSpecifiers = config.extends ? [config.extends].flat().map(resolveExtendSpecifier) : [];
    if (extendsSpecifiers.some(specifier => specifier?.startsWith('eslint-plugin-prettier')))
        extendsSpecifiers.push('eslint-config-prettier');
    const plugins = config.plugins ? config.plugins.map(resolvePluginSpecifier) : [];
    const parser = config.parser ?? config.parserOptions?.parser;
    const babelDependencies = config.parserOptions?.babelOptions
        ? getDependenciesFromConfig(config.parserOptions.babelOptions)
        : [];
    const settings = config.settings ? getDependenciesFromSettings(config.settings) : [];
    const overrides = config.overrides ? [config.overrides].flat().flatMap(getDependencies) : [];
    return compact([...extendsSpecifiers, ...plugins, parser, ...babelDependencies, ...settings, ...overrides]);
};
export const getDependenciesDeep = async (configFilePath, options, dependencies = new Set()) => {
    const addAll = (deps) => deps.forEach(dependency => dependencies.add(dependency));
    const localConfig = basename(configFilePath) === 'package.json'
        ? options.manifest[PACKAGE_JSON_PATH]
        : /(\.(jsonc?|ya?ml)|rc)$/.test(configFilePath)
            ? await load(configFilePath)
            : await fallback(configFilePath);
    if (localConfig) {
        if (localConfig.extends) {
            for (const extend of [localConfig.extends].flat()) {
                if (isInternal(extend)) {
                    const filePath = toAbsolute(extend, dirname(configFilePath));
                    const extendConfigFilePath = _resolve(filePath);
                    dependencies.add(extendConfigFilePath);
                    addAll(await getDependenciesDeep(extendConfigFilePath, options, dependencies));
                }
            }
        }
        addAll(getDependencies(localConfig));
    }
    return dependencies;
};
const isQualifiedSpecifier = (specifier) => specifier === 'eslint' ||
    /\/eslint-(config|plugin)$/.test(specifier) ||
    /.+eslint-(config|plugin)\//.test(specifier) ||
    /eslint-(config|plugin)-/.test(specifier);
const resolveSpecifier = (namespace, rawSpecifier) => {
    const specifier = rawSpecifier.replace(/(^plugin:|:.+$)/, '');
    if (isQualifiedSpecifier(specifier))
        return specifier;
    if (!specifier.startsWith('@')) {
        const id = rawSpecifier.startsWith('plugin:') ? getPackageNameFromModuleSpecifier(specifier) : specifier;
        return `${namespace}-${id}`;
    }
    const [scope, name, ...rest] = specifier.split('/');
    if (rawSpecifier.startsWith('plugin:') && rest.length === 0)
        return [scope, namespace].join('/');
    return [scope, name ? `${namespace}-${name}` : namespace, ...rest].join('/');
};
const resolvePluginSpecifier = (specifier) => resolveSpecifier('eslint-plugin', specifier);
const resolveExtendSpecifier = (specifier) => {
    if (isInternal(specifier))
        return;
    if (/^next(\/.+)?$/.test(specifier))
        return specifier;
    const namespace = specifier.startsWith('plugin:') ? 'eslint-plugin' : 'eslint-config';
    return resolveSpecifier(namespace, specifier);
};
const getDependenciesFromSettings = (settings = {}) => {
    return Object.entries(settings).flatMap(([settingKey, settings]) => {
        if (settingKey === 'import/resolver') {
            return (typeof settings === 'string' ? [settings] : Object.keys(settings))
                .filter(key => key !== 'node')
                .map(key => {
                if (isInternal(key))
                    return key;
                if (isAbsolute(key))
                    return getPackageNameFromFilePath(key);
                return `eslint-import-resolver-${key}`;
            });
        }
        if (settingKey === 'import/parsers') {
            return (typeof settings === 'string' ? [settings] : Object.keys(settings)).map(key => {
                if (isAbsolute(key))
                    return getPackageNameFromFilePath(key);
                return key;
            });
        }
    });
};
