import { isBuiltin } from 'node:module';
import ts from 'typescript';
import { getOrSet } from '../util/map.js';
import { isStartsLikePackageName, sanitizeSpecifier } from '../util/modules.js';
import { isInNodeModules } from '../util/path.js';
import { timerify } from '../util/Performance.js';
import { isDeclarationFileExtension, isAccessExpression, getAccessExpressionName, getJSDocTags, getLineAndCharacterOfPosition, } from './ast-helpers.js';
import getExportVisitors from './visitors/exports/index.js';
import { getJSXImplicitImportBase } from './visitors/helpers.js';
import getImportVisitors from './visitors/imports/index.js';
import getScriptVisitors from './visitors/scripts/index.js';
const getVisitors = (sourceFile) => ({
    export: getExportVisitors(sourceFile),
    import: getImportVisitors(sourceFile),
    script: getScriptVisitors(sourceFile),
});
const getImportsAndExports = (sourceFile, getResolvedModule, options) => {
    const internalImports = new Map();
    const externalImports = new Set();
    const unresolvedImports = new Set();
    const exports = new Map();
    const aliasedExports = new Map();
    const scripts = new Set();
    const importedInternalSymbols = new Map();
    const jsxImport = getJSXImplicitImportBase(sourceFile);
    if (jsxImport)
        externalImports.add(jsxImport);
    const visitors = getVisitors(sourceFile);
    const addInternalImport = (options) => {
        const { identifier, specifier, symbol, filePath, isReExport } = options;
        const isStar = identifier === '*';
        const internalImport = getOrSet(internalImports, filePath, {
            specifier,
            hasStar: isStar,
            isReExport,
            isReExportedBy: new Set(),
            symbols: new Set(),
        });
        if (isReExport) {
            internalImport.isReExport = isReExport;
            internalImport.isReExportedBy.add(sourceFile.fileName);
        }
        if (isStar)
            internalImport.hasStar = true;
        if (!isStar)
            internalImport.symbols.add(identifier);
        if (isStar && symbol)
            importedInternalSymbols.set(symbol, filePath);
    };
    const addImport = (options) => {
        const { specifier, symbol, identifier = '__anonymous', isReExport = false, pos } = options;
        if (isBuiltin(specifier))
            return;
        const module = getResolvedModule(specifier);
        if (module?.resolvedModule) {
            const filePath = module.resolvedModule.resolvedFileName;
            if (filePath) {
                if (module.resolvedModule.isExternalLibraryImport) {
                    if (!isInNodeModules(filePath)) {
                        addInternalImport({ identifier, specifier, symbol, filePath, isReExport });
                    }
                    const sanitizedSpecifier = sanitizeSpecifier(specifier);
                    if (!isStartsLikePackageName(sanitizedSpecifier)) {
                        return;
                    }
                    if (isDeclarationFileExtension(module.resolvedModule.extension)) {
                        externalImports.add(sanitizedSpecifier);
                    }
                    else {
                        externalImports.add(module.resolvedModule.packageId?.name ?? sanitizedSpecifier);
                    }
                }
                else {
                    addInternalImport({ identifier, specifier, symbol, filePath, isReExport });
                }
            }
        }
        else {
            if (typeof pos === 'number') {
                const { line, character } = sourceFile.getLineAndCharacterOfPosition(pos);
                unresolvedImports.add({ specifier, pos, line: line + 1, col: character + 1 });
            }
            else {
                unresolvedImports.add({ specifier });
            }
        }
    };
    const maybeAddNamespaceAccessAsImport = ({ namespace, member }) => {
        const symbol = sourceFile.locals?.get(namespace);
        if (symbol) {
            const importedSymbolFilePath = importedInternalSymbols.get(symbol);
            if (importedSymbolFilePath) {
                const internalImport = internalImports.get(importedSymbolFilePath);
                internalImport?.symbols.add(member);
            }
        }
    };
    const addExport = ({ node, identifier, type, pos, posDecl, members = [], fix }) => {
        if (options.skipExports)
            return;
        const jsDocTags = getJSDocTags(node);
        if (exports.has(identifier)) {
            const item = exports.get(identifier);
            const crew = [...item.members, ...members];
            const tags = new Set([...item.jsDocTags, ...jsDocTags]);
            exports.set(identifier, { ...item, members: crew, jsDocTags: tags });
        }
        else {
            exports.set(identifier, { node, type, members, jsDocTags, pos, posDecl: posDecl ?? pos, fix });
        }
        if (!jsDocTags.has('@alias')) {
            if (ts.isExportAssignment(node))
                maybeAddAliasedExport(node.expression, 'default');
            if (ts.isVariableDeclaration(node))
                maybeAddAliasedExport(node.initializer, identifier);
        }
    };
    const maybeAddAliasedExport = (node, alias) => {
        const identifier = node?.getText();
        if (node && identifier) {
            const exprt = sourceFile.symbol?.exports?.get(identifier);
            if (exprt && exprt.valueDeclaration) {
                if (!aliasedExports.has(identifier)) {
                    const pos = getLineAndCharacterOfPosition(exprt.valueDeclaration, exprt.valueDeclaration.pos);
                    aliasedExports.set(identifier, [{ symbol: identifier, ...pos }]);
                }
                const i = aliasedExports.get(identifier);
                const pos = getLineAndCharacterOfPosition(node, node.pos);
                i?.push({ symbol: alias, ...pos });
            }
        }
    };
    const visit = (node) => {
        for (const visitor of visitors.import) {
            if (visitor) {
                const results = visitor(node, options);
                if (results)
                    [results].flat().forEach(addImport);
            }
        }
        const isTopLevel = node.parent === sourceFile || node.parent?.parent === sourceFile;
        if (isTopLevel) {
            for (const visitor of visitors.export) {
                if (visitor) {
                    const results = visitor(node, options);
                    if (results)
                        [results].flat().forEach(addExport);
                }
            }
        }
        for (const visitor of visitors.script) {
            if (visitor) {
                const results = visitor(node, options);
                if (results)
                    [results].flat().forEach(script => scripts.add(script));
            }
        }
        if (isAccessExpression(node)) {
            maybeAddNamespaceAccessAsImport({
                namespace: node.expression.getText(),
                member: getAccessExpressionName(node),
            });
        }
        ts.forEachChild(node, visit);
    };
    visit(sourceFile);
    return {
        imports: {
            internal: internalImports,
            external: externalImports,
            unresolved: unresolvedImports,
        },
        exports: {
            exported: exports,
            duplicate: [...aliasedExports.values()],
        },
        scripts,
    };
};
export const _getImportsAndExports = timerify(getImportsAndExports);
